/* Copyright (C) 2004   db4objects Inc.   http://www.db4o.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA  02111-1307, USA. */

package com.db4o.ext;

/**
 * carries in-memory data for db4o in-memory operation.
 * <br><br>In-memory ObjectContainers are useful for maximum performance
 * on small databases, for swapping objects or for storing db4o format data
 * to other media or other databases.<br><br>Be aware of the danger of running
 * into OutOfMemory problems or complete loss of all data, in case of hardware
 * or JVM failures.
 * <br><br>
 * @see <a href="ExtDb4o.html#openMemoryFile(com.db4o.MemoryFile)">
 * <code>ExtDb4o#openMemoryFile()</code></a>
 */
public class MemoryFile {

	private byte[] i_bytes;
	private final int INITIAL_SIZE_AND_INC = 10000;
	private int i_initialSize = INITIAL_SIZE_AND_INC;
	private int i_incrementSizeBy = INITIAL_SIZE_AND_INC;
	

	/**
	 * constructs a new MemoryFile without any data.
	 * @see <a href="ExtDb4o.html#openMemoryFile(com.db4o.MemoryFile)"> 
	 * <code>ExtDb4o#openMemoryFile()</code></a>
	 */
	public MemoryFile() {
	}

	/**
	 * constructs a MemoryFile to use the byte data from a previous
	 * MemoryFile.
	 * @param bytes the raw byte data.
	 * @see <a href="ExtDb4o.html#openMemoryFile(com.db4o.MemoryFile)">
	 * <code>ExtDb4o#openMemoryFile()</code></a>
	 */
	public MemoryFile(byte[] bytes) {
		i_bytes = bytes;
	}
	
	/**
	 * returns the raw byte data.
	 * <br><br>Use this method to get the byte data from the MemoryFile
	 * to store it to other media or databases, for backup purposes or
	 * to create other MemoryFile sessions.
	 * <br><br>The byte data from a MemoryFile should only be used
	 * after it is closed.<br><br>
	 * @return bytes the raw byte data.
	 */
	public byte[] getBytes(){
		if(i_bytes == null){
			return new byte[0];
		}
		return i_bytes;
	}
	
	/**
	 * returns the size the MemoryFile is to be enlarged, if it grows beyond
	 * the current size.
	 * @return size in bytes
	 */
	public int getIncrementSizeBy(){
		return i_incrementSizeBy;
	}
	
	/**
	 * returns the initial size of the MemoryFile.
	 * @return size in bytes
	 */
	public int getInitialSize(){
		return i_initialSize;
	}
	
	/**
	 * sets the raw byte data.
	 * <br><br><b>Caution!</b><br>Calling this method during a running
	 * Memory File session may produce unpreditable results.
	 * @param bytes the raw byte data.
	 */
	public void setBytes(byte[] bytes){
		i_bytes = bytes;
	}

	/**
	 * configures the size the MemoryFile is to be enlarged by, if it grows
	 * beyond the current size.
	 * <br><br>Call this method before passing the MemoryFile to 
	 * <a href="ExtDb4o.html#openMemoryFile(com.db4o.MemoryFile)">
	 * <code>ExtDb4o#openMemoryFile(MemoryFile)</code></a>.<br><br>
	 * This parameter can be modified to tune the maximum performance of
	 * a MemoryFile for a specific usecase. To produce the best results,
	 * test the speed of your application with real data.<br><br>
	 * @param byteCount the desired size in bytes
	 */
	public void setIncrementSizeBy(int byteCount){
		i_incrementSizeBy = byteCount;
	}
	
	/**
	 * configures the initial size of the MemoryFile.
	 * <br><br>Call this method before passing the MemoryFile to 
	 * <a href="ExtDb4o.html#openMemoryFile(com.db4o.MemoryFile)">
	 * <code>ExtDb4o#openMemoryFile(MemoryFile)</code></a>.<br><br>
	 * This parameter can be modified to tune the maximum performance of
	 * a MemoryFile for a specific usecase. To produce the best results,
	 * test speed and memory consumption of your application with
	 * real data.<br><br>
	 * @param byteCount the desired size in bytes
	 */
	public void setInitialSize(int byteCount){
		i_initialSize = byteCount;
	}
}